'''
PyFerret Python External Function (PyEF) for regridding data from
a curvilinear longitude, latitude grid to a rectilinear longitude,
latitude grid.  Uses the Ferret2DRegridder class, which in turn
uses ESMP/ESMF, to perform the regridding.

@author: Karl Smith
'''
import numpy
import pyferret
import ESMP
import pyferret.regrid as regrid


def ferret_init(efid):
    '''
    Initializes the curv2rect function.  Either center or corner curvilinear
    coordinates must be given; both may be given.  If only corner coordinates
    are given, center coordinates are positioned at the centroid of each cell
    as computed using the Surveyor's Formula.  If conservative regridding is
    used, corners must be given.
    '''
    init_dict = { }
    init_dict["numargs"] = 7
    init_dict["descript"] = \
        "Regrids data from curvilinear lon,lat center " \
        "and/or corner grid to rectilinear using ESMP/ESMF"
    init_dict["argnames"] = ("CurvData",
                             "CurvCenterLons",
                             "CurvCenterLats",
                             "CurvCornerLons",
                             "CurvCornerLats",
                             "TemplateRectVar",
                             "Method")
    init_dict["argdescripts"] = (
        "Curvilinear X,Y data positioned at centers",
        "Curvilinear center longitudes on an X,Y grid",
        "Curvilinear center latitudes on an X,Y grid",
        "Curvilinear corner longitudes on an X,Y (maybe Z) grid",
        "Curvilinear corner latitudes on an X,Y (maybe Z) grid",
        "Template variable on the desired rectilinear grid",
        "Regrid method: BILINEAR, PATCH, CONSERVE")
    init_dict["argtypes"] = (pyferret.FLOAT_ARRAY,
                             pyferret.FLOAT_ARRAY,
                             pyferret.FLOAT_ARRAY,
                             pyferret.FLOAT_ARRAY,
                             pyferret.FLOAT_ARRAY,
                             pyferret.FLOAT_ARRAY,
                             pyferret.STRING_ONEVAL)
    no_influence = [ False ] * pyferret.MAX_FERRET_NDIM
    full_influence = [ True ] * pyferret.MAX_FERRET_NDIM
    part_influence = full_influence[:]
    part_influence[pyferret.X_AXIS] = False
    part_influence[pyferret.Y_AXIS] = False
    init_dict["influences"] = (part_influence,
                               no_influence,
                               no_influence,
                               no_influence,
                               no_influence,
                               full_influence,
                               no_influence)
    # okay to break up along Z, T, E, and F result axes
    init_dict["piecemeal"] = (False, False, True, True, True, True)

    return init_dict


def ferret_compute(efid, result, result_bdf, inputs, input_bdfs):
    '''
    Performs the regridding for the curv2rect function.

    Arguments:
        result     - numpy array for rectilinear data values to be assigned
        result_bdf - numpy array of the missing-data value for result
        inputs     - numpy arrays (CurvData,
                                   CurvCenterLons, CurvCenterLats,
                                   CurvCornerLons, CurvCornerLats.
                                   TemplateRectVar, Method)
        input_bdfs - numpy arrays of missing-data values for the
                     corresponding inputs array
    '''
    # Get the regridding method to use
    methodstr = pyferret.get_arg_one_val(efid, pyferret.ARG7).upper()
    if methodstr == "BILINEAR":
        method = ESMP.ESMP_REGRIDMETHOD_BILINEAR
    elif methodstr == "PATCH":
        method = ESMP.ESMP_REGRIDMETHOD_PATCH
    elif methodstr == "CONSERVE":
        method = ESMP.ESMP_REGRIDMETHOD_CONSERVE
    else:
        raise ValueError("Unknown method %s" % methodstr)

    # Get the template data and missing value
    template_data = inputs[pyferret.ARG6]
    template_undef = input_bdfs[pyferret.ARG6]

    # Get the rectilinear center longitudes and latitudes
    rect_center_lons = pyferret.get_axis_coordinates(efid,
                                    pyferret.ARG6, pyferret.X_AXIS)
    rect_center_lats = pyferret.get_axis_coordinates(efid,
                                    pyferret.ARG6, pyferret.Y_AXIS)

    # Get the rectilinear corner longitudes
    lo, hi = pyferret.get_axis_box_limits(efid, pyferret.ARG6, pyferret.X_AXIS)
    if not numpy.allclose(lo[1:], hi[:-1]):
        raise ValueError("Unexpected ARG6 X_AXIS box limit values " \
                         "returned from pyferret.get_axis_box_limits")
    rect_corner_lons = numpy.empty((lo.shape[0] + 1), dtype=numpy.float64)
    rect_corner_lons[0] = lo[0]
    rect_corner_lons[1:] = hi

    # Get the rectilinear corner latitudes
    lo, hi = pyferret.get_axis_box_limits(efid, pyferret.ARG6, pyferret.Y_AXIS)
    if not numpy.allclose(lo[1:], hi[:-1]):
        raise ValueError("Unexpected ARG6 Y_AXIS box limit values " \
                         "returned from pyferret.get_axis_box_limits")
    rect_corner_lats = numpy.empty((lo.shape[0] + 1), dtype=numpy.float64)
    rect_corner_lats[0] = lo[0]
    rect_corner_lats[1:] = hi

    # Get the curvilinear data
    curv_data = inputs[pyferret.ARG1]
    curv_centers_shape = (curv_data.shape[0], curv_data.shape[1])
    if (curv_data.shape[2:] != template_data.shape[2:]):
        raise ValueError("Curvilinear data and template variable " \
                         "must have same Z, T, E, and F axes")
    curv_undef = input_bdfs[pyferret.ARG1]

    # Get the curvilinear centers arrays
    lons = inputs[pyferret.ARG2].squeeze()
    lats = inputs[pyferret.ARG3].squeeze()
    if (lons.shape == curv_centers_shape) and (lats.shape == curv_centers_shape):
        curv_center_lons = lons
        curv_center_lats = lats
    elif (lons.shape == ()) and (lats.shape == ()):
        curv_center_lons = None
        curv_center_lats = None
    else:
        raise ValueError("Curvilinear center points must have " \
                         "appropriate shape or be singletons")

    # Get the curvilinear corners arrays
    lons = inputs[pyferret.ARG4].squeeze()
    lats = inputs[pyferret.ARG5].squeeze()
    corners_shape = (curv_data.shape[0] + 1, curv_data.shape[1] + 1)
    corners_shape_3d = (curv_data.shape[0], curv_data.shape[1], 4)
    if (lons.shape == corners_shape) and (lats.shape == corners_shape):
        curv_corner_lons = lons
        curv_corner_lats = lats
    elif (lons.shape == corners_shape_3d) and (lats.shape == corners_shape_3d):
        curv_corner_lons, curv_corner_lats = regrid.quadCornersFrom3D(lons, lats)
    elif method == ESMP.ESMP_REGRIDMETHOD_CONSERVE:
        raise ValueError("CONSERVE method requires " \
                         "curvilinear grid corner coordinates")
    elif (lons.shape == ()) and (lats.shape == ()):
        curv_corner_lons = None
        curv_corner_lats = None
    else:
        raise ValueError("Curvilinear corner points must have " \
                         "appropriate shape or be singletons")

    # If curvilinear center point coordinates not given, 
    # generate them from the curvilinear corner points
    if curv_center_lons is None:
        if not curv_corner_lons is None:
            curv_center_lons, curv_center_lats = \
                regrid.quadCentroids(curv_corner_lons, curv_corner_lats)
        else:
            raise ValueError("Valid center or corner curvilinear " \
                             "grid coordinates must be given")

    # Make sure ESMP is, or has been, initialized
    regrid.ESMPControl().startCheckESMP()

    # Create the regridder used repeatedly in this function 
    regridder = regrid.CurvRectRegridder()
    last_curv_center_ignore = None
    last_rect_center_ignore = None

    # Increment the depth index last
    # most likely to change the undefined (e.g., land) mask 
    for d_idx in range(curv_data.shape[2]):
        # Arrays are probably in Fortran order, so increment last indices last
        for f_idx in range(curv_data.shape[5]):
            for e_idx in range(curv_data.shape[4]):
                for t_idx in range(curv_data.shape[3]):
                    # Determine curvilinear center points to ignore from undefined data
                    curv_center_ignore = ( numpy.abs(curv_data[:, :, d_idx, t_idx,
                                                e_idx, f_idx] - curv_undef) < 1.0E-7 )
                    # If mask has changed, need to recreate the curvilinear grid
                    if (last_curv_center_ignore is None) or \
                       numpy.any(curv_center_ignore != last_curv_center_ignore):
                        regridder.createCurvGrid(curv_center_lons, curv_center_lats,
                                                 curv_center_ignore, curv_corner_lons,
                                                 curv_corner_lats, None)
                        last_curv_center_ignore = curv_center_ignore
                    # Determine rectilinear center points to ignore from undefined data
                    rect_center_ignore = ( numpy.abs(template_data[:, :, d_idx, t_idx,
                                                e_idx, f_idx] - template_undef) < 1.0E-7 )
                    # If mask has changed, need to recreate the rectilinear grid
                    if (last_rect_center_ignore is None) or \
                       numpy.any(rect_center_ignore != last_rect_center_ignore):
                        regridder.createRectGrid(rect_center_lons, rect_center_lats,
                                                 rect_center_ignore, rect_corner_lons,
                                                 rect_corner_lats, None)
                        last_rect_center_ignore = rect_center_ignore
                    # Assign the curvilinear data
                    regridder.assignCurvField(curv_data[:, :, d_idx, t_idx, e_idx, f_idx])
                    # Allocate space for the rectilinear data from the regridding
                    regridder.assignRectField(None)
                    # Regrid and assign the rectilinear data to the results array
                    regrid_data = regridder.regridCurvToRect(result_bdf, method)
                    result[:, :, d_idx, t_idx, e_idx, f_idx] = regrid_data

    return




'''
Template for creating a PyFerret Python External Function (PyEF).
The names of the functions provided should not be changed.  By 
default, PyFerret uses the name of the module as the function name.

Copy this file using a name that you would like to be the function
name, then modify the contents of these functions and comments as
desired. 
'''

import numpy


def ferret_init(efid):
    '''
    Initialization function for this PyFerret PyEF.  Returns
    a dictionary describing the features of this PyFerret PyEF.
    At a minimum, assigns the number of arguments expected and 
    a descriptions of the functions.  May also provide 
    descriptions of the arguments and specifications for a 
    non-standard result grid.
    '''
    init_dict = { }

    init_dict["numargs"] = 1
    init_dict["descript"] = "Pass through"

    return init_dict


def ferret_result_limits(efid):
    '''
    Defines the index limits for all abstract axes in the result grid.
    Returns an (X,Y,Z,T,E,F)-axis six-tuple of either (low,high) pairs, 
    for an abstract axis, or None, for a non-abstract axis.  The low 
    and high values are integer values.  If the result grid has no 
    abstract axes, this function will not be called and can be deleted.
    '''
    axis_limits = (None, None, None, None, None, None)

    return axis_limits


def ferret_custom_axes(efid):
    '''
    Defines all custom axes in ther result grid.  Returns a (X,Y,Z,T,E,F)-
    axis six-tuple of either a (low, high, delta, unit_name, is_modulo) 
    tuple, for a custom axis, or None, for a non-custom axis.  The low,
    high, and delta values are floating-point values in units of the axis 
    coordinate ("world coordinates").  If the result grid has no custom
    axes, this function will not be called and can be deleted.
    '''
    axis_info = (None, None, None, None, None, None)
    
    return axis_info


def ferret_compute(efid, result, result_bdf, inputs, input_bdfs):
    '''
    Computation function for this PyFerret PyEF.  Assign values to the
    elements of result; do not reassign result itself.  In other words,
    assign values using notation similar to 'result[...] = ...'; do not
    use notation similar to 'result = ...' as this will simply define
    a new local variable called result, hiding the variable passed into
    this function.

    If an error is detected, raise an appropriate exception.  ValueError
    is commonly used for unexpected values.  IndexError is commonly used 
    for unexpected array sizes.

    Arguments:
        result     - numpy float array to be assigned
        result_bdf - numpy read-only float array of one element giving the 
                     missing-data value to be used when assigning result
        inputs     - tuple of numpy read-only float arrays giving the input
                     values provided by the caller
        input_bdfs - numpy read-only float arrays of one element giving the
                     missing-data value for the corresponding inputs array
    '''
    # Create masks of values that are undefined and that are defined
    bad_mask = ( inputs[0] == input_bdfs[0] )
    good_mask = numpy.logical_not(bad_mask)
    result[good_mask] = inputs[0][good_mask]
    result[bad_mask] = result_bdf
    return




"""
Performs a two-sided Kolmogorov-Smirnov test that the provided
sample comes from the given probability distribution function.
"""

from __future__ import print_function

import numpy
import pyferret
import pyferret.stats
import scipy.stats


def ferret_init(id):
    """
    Initialization for the stats_kstest1 PyEF
    """
    axes_values = [ pyferret.AXIS_DOES_NOT_EXIST ] * pyferret.MAX_FERRET_NDIM
    axes_values[0] = pyferret.AXIS_CUSTOM
    false_influences = [ False ] * pyferret.MAX_FERRET_NDIM
    retdict = { "numargs": 3,
                "descript": "Returns two-sided Kolmogorov-Smirnov test stat. and prob. " \
                            "that sample comes from a pop. with given prob. distrib.",
                "axes": axes_values,
                "argnames": ( "SAMPLE", "PDNAME", "PDPARAMS", ),
                "argdescripts": ( "Sample data array",
                                  "Name of a continuous probability distribution",
                                  "Parameters for this continuous probability distribution"),
                "argtypes": ( pyferret.FLOAT_ARRAY, pyferret.STRING_ONEVAL, pyferret.FLOAT_ARRAY, ),
                "influences": ( false_influences, false_influences, false_influences, ),
              }
    return retdict


def ferret_custom_axes(id):
    """
    Define custom axis of the stats_kstest1 Ferret PyEF
    """
    axis_defs = [ None ] * pyferret.MAX_FERRET_NDIM
    axis_defs[0] = ( 1, 2, 1, "KS,P", False )
    return axis_defs


def ferret_compute(id, result, resbdf, inputs, inpbdfs):
    """
    Performs a two-sided Kolmogorov-Smirnov test that the provided sample
    comes from a population with the given probability distribution function.
    The sample is given in inputs[0], the probability distribution function
    name is given in inputs[1] (a string), and the "standard" parameters for
    this probability distribution function are given in inputs[2].  The test
    statistic value and two-tailed probability are returned in result.
    Undefined data given in inputs[0] are removed before performing the test.
    """
    # get the scipy.stats distribution name from the given distribution name
    if inputs[1] is None:
        raise ValueError("The name of a probability distribution function not given")
    distscipyname = pyferret.stats.getdistname(inputs[1])
    if distscipyname is None:
        raise ValueError("Unknown or unsupported probability distribution function %s" % inputs[1])
    # get the scipy.stats distribution parameters from the given "standard" parameters
    if inputs[2] is None:
        raise ValueError("Paramaters for the probability distribution function not given")
    distscipyparams = pyferret.stats.getdistparams(distscipyname, inputs[2].reshape(-1))
    if distscipyparams is None:
        raise ValueError("Unknown or unsupported (for params) probability distribution function %s" % inputs[1])
    # get the valid sample values
    badmask = ( numpy.fabs(inputs[0] - inpbdfs[0]) < 1.0E-5 )
    badmask = numpy.logical_or(badmask, numpy.isnan(inputs[0]))
    goodmask = numpy.logical_not(badmask)
    values = inputs[0][goodmask]
    # perform the test and assign the results
    fitparams = scipy.stats.kstest(values, distscipyname, distscipyparams)
    result[:] = resbdf
    # Kolmogorov-Smirnov test statistic
    result[0] = fitparams[0]
    # probability
    result[1] = fitparams[1]


#
# The rest of this is just for testing this module at the command line
#
if __name__ == "__main__":
    # make sure ferret_init and ferret_custom_axes do not have problems
    info = ferret_init(0)
    info = ferret_custom_axes(0)

    # Set the seed to reproduce a problematic distribution
    # import numpy.random
    # numpy.random.seed(3333333)

    # Get a random sample from the compared distribution and from another distribution
    ydim = 200
    zdim = 150
    mu = 5.0
    sigma = 0.5
    rvsc = scipy.stats.norm(mu, sigma).rvs(ydim * zdim)
    rvsu = scipy.stats.uniform(loc=(mu + 3.0 * sigma), scale=(3.0 * sigma)).rvs(ydim * zdim)

    # setup for the call to ferret_compute
    distname = "norm"
    distparams = numpy.array([mu, sigma], dtype=numpy.float64)
    inpbdfs = numpy.array([-9999.0, -1.0, -2.0], dtype=numpy.float64)
    resbdf  = numpy.array([-8888.0], dtype=numpy.float64)
    sampc = numpy.empty((1, ydim, zdim, 1, 1, 1), dtype=numpy.float64, order='F')
    sampu = numpy.empty((1, ydim, zdim, 1, 1, 1), dtype=numpy.float64, order='F')
    index = 0
    for j in range(ydim):
        for k in range(zdim):
            if (index % 71) == 3:
                sampc[0, j, k, 0, 0, 0] = inpbdfs[0]
                sampu[0, j, k, 0, 0, 0] = inpbdfs[0]
            else:
                sampc[0, j, k, 0, 0, 0] = rvsc[index]
                sampu[0, j, k, 0, 0, 0] = rvsu[index]
            index += 1
    resultc = -7777.0 * numpy.ones((2, 1, 1, 1, 1, 1), dtype=numpy.float64, order='F')
    resultu = -7777.0 * numpy.ones((2, 1, 1, 1, 1, 1), dtype=numpy.float64, order='F')

    # call ferret_compute with data from the distribution and check the results
    ferret_compute(0, resultc, resbdf, (sampc, distname, distparams), inpbdfs)
    resultc = resultc.reshape(-1)
    print("from same dist result: %s" % str(resultc))
    if (resultc[0] < 0.00) or (resultc[0] > 0.01) or \
       (resultc[1] < 0.10) or (resultc[1] > 1.00):
        raise ValueError("Unexpected result")

    # call ferret_compute with data from a different distribution and check the results
    ferret_compute(0, resultu, resbdf, (sampu, distname, distparams), inpbdfs)
    resultu = resultu.reshape(-1)
    print("from diff dist result:  %s" % str(resultu))
    if (resultu[0] < 0.99) or (resultu[0] > 1.00) or \
       (resultu[1] < 0.00) or (resultu[1] > 0.01):
        raise ValueError("Unexpected result")

    # All successful
    print("Success")




# To be run in python after importing and starting pyferret
# such as from running "pyferret -python"

from __future__ import print_function

import sys ; sys.ps1 = '' ; sys.ps2 = ''
print()

print(">>> coads = pyferret.FerDSet('coads_climatology.cdf')")
coads = pyferret.FerDSet('coads_climatology.cdf')
print(">>> pyferret.showdata(brief=False)")
pyferret.showdata(brief=False)

print(">>> coads.sst.load()")
coads.sst.load()
print(">>> datacopy = coads.sst.data")
datacopy = coads.sst.data
print(">>> datacopy[87:93,43:47,0,:,0,0] = -5.0")
datacopy[87:93,43:47,0,:,0,0] = -5.0
print(">>> coads.sstcopy = pyferret.FerPyVar(grid=coads.sst.grid, data=datacopy, missval=coads.sst.missval, unit=coads.sst.unit, title='modified copy of SST')")
coads.sstcopy = pyferret.FerPyVar(grid=coads.sst.grid, data=datacopy, missval=coads.sst.missval, unit=coads.sst.unit, title='modified copy of SST')
print(">>> pyferret.showdata(brief=False)")
pyferret.showdata(brief=False)
print(">>> pyferret.setwindow(1, axisasp=0.5, logo=False)")
pyferret.setwindow(1,axisasp=0.5,logo=False)
print(">>> pyferret.setdefaulttext(font='helvetica')")
pyferret.settextstyle(font='helvetica')
print(">>> pyferret.shadeplot(coads.sstcopy['40S':'40N','100E':'100W','16-FEB'])")
pyferret.shadeplot(coads.sstcopy['40S':'40N','100E':'100W','16-FEB'])
print(">>> pyferret.saveplot('testferpyvar.pdf')")
pyferret.saveplot('testferpyvar.pdf')

# print ">>> pyferret.run('let sstdiff = sst - sstcopy')"
# pyferret.run('let sstdiff = sst - sstcopy')
# print ">>> pyferret.run('vtree sstdiff')"
# pyferret.run('vtree sstdiff')

print(">>> del coads.sstcopy")
del coads.sstcopy
print(">>> pyferret.showdata()")
pyferret.showdata()
print(">>> coads.close()")
coads.close()
print(">>> pyferret.showdata()")
pyferret.showdata()

